# Copyright 2023 Agnostiq Inc.


from functools import partial
from pathlib import Path
from typing import Dict

import toml

from ..shared.classes.settings import Settings


class AuthConfigManager:
    @staticmethod
    def get_config_file(settings: Settings = None) -> str:
        """
        Returns the path to the config file.

        Args:
            settings: The settings object to use. If None, the default settings will be used.
        
        Returns:
            The path to the config file.

        """

        if settings is None:
            settings = Settings()

        path = Path(settings.auth.config_file)
        path.parent.mkdir(parents=True, exist_ok=True)
        return str(path.resolve())

    @staticmethod
    def get_auth_request_headers(settings: Settings = None) -> Dict[str, str]:
        """
        Returns the headers to use for an authentication request.

        Args:
            settings: The settings object to use. If None, the default settings will be used.
        
        Returns:
            Authentication request headers as a dictionary.

        """

        if settings is None:
            settings = Settings()
        return {"x-api-key": f"{AuthConfigManager.get_api_key(settings)}"}

    @staticmethod
    def save_token(token: str, settings: Settings = None) -> None:
        """
        Saves the authentication token to the config file.

        Args:
            token: The authentication token to save.
            settings: The settings object to use. If None, the default settings will be used.
        
        Returns:
            None

        """

        if settings is None:
            settings = Settings()

        auth_section_header = settings.auth.config_file_section
        token_keyname = settings.auth.cofig_file_token_keyname
        
        toml_dict = {auth_section_header: {}}
        toml_dict[auth_section_header][token_keyname] = token or ""
        
        with open(AuthConfigManager.get_config_file(settings), "w") as f:
            toml.dump(toml_dict, f)

    @staticmethod
    def get_token(settings: Settings = None) -> str:
        """
        Returns the authentication token from the config file.

        Args:
            settings: The settings object to use. If None, the default settings will be used.
        
        Returns:
            The authentication token.

        """

        if settings is None:
            settings = Settings()

        token = settings.auth.token
        if not token:
            auth_section_header = settings.auth.config_file_section
            token_keyname = settings.auth.cofig_file_token_keyname
            
            with open(AuthConfigManager.get_config_file(settings), "r") as f:
                toml_string = f.read()
                parsed_toml = toml.loads(toml_string)
                token = parsed_toml[auth_section_header][token_keyname]
        return token

    @staticmethod
    def get_api_key(settings: Settings = None) -> str:
        """
        Returns the API key from the config file.

        Args:
            settings: The settings object to use. If None, the default settings will be used.
        
        Returns:
            The API key.

        """

        if settings is None:
            settings = Settings()

        api_key = settings.auth.api_key
        if not api_key:
            auth_section_header = settings.auth.config_file_section
            api_keyname = settings.auth.config_file_api_key_keyname

            with open(AuthConfigManager.get_config_file(settings), "r") as f:
                toml_string = f.read()
                parsed_toml = toml.loads(toml_string)
                api_key = parsed_toml[auth_section_header][api_keyname]
        return api_key

    @staticmethod
    def save_api_key(api_key: str, settings: Settings = None) -> None:
        """
        Saves the API key to the config file.

        Args:
            api_key: The API key to save.
            settings: The settings object to use. If None, the default settings will be used.
        
        Returns:
            None

        """

        if settings is None:
            settings = Settings()

        auth_section_header = settings.auth.config_file_section

        api_keyname = settings.auth.config_file_api_key_keyname

        toml_dict = {auth_section_header: {}}
        toml_dict[auth_section_header][api_keyname] = api_key or ""

        with open(AuthConfigManager.get_config_file(settings), "w") as f:
            toml.dump(toml_dict, f)


get_token = partial(AuthConfigManager.get_token)
save_token = partial(AuthConfigManager.save_token)

get_api_key = partial(AuthConfigManager.get_api_key)
save_api_key = partial(AuthConfigManager.save_api_key)



# Copyright 2023 Agnostiq Inc.

"""Functions to convert node -> ElectronSchema"""


from covalent._shared_files.util_classes import RESULT_STATUS, Status

from ..shared.schemas.electron import (
    ASSET_FILENAME_MAP,
    ElectronAssets,
    ElectronMetadata,
    ElectronSchema,
)
from .common import AssetType, load_asset, save_asset

ASSET_TYPES = {
    "function": AssetType.TRANSPORTABLE,
    "function_string": AssetType.TEXT,
    "value": AssetType.TRANSPORTABLE,
    "output": AssetType.TRANSPORTABLE,
    "deps": AssetType.OBJECT,
    "call_before": AssetType.OBJECT,
    "call_after": AssetType.OBJECT,
    "stdout": AssetType.TEXT,
    "stderr": AssetType.TEXT,
    "error": AssetType.TEXT,
}


def _serialize_node_metadata(node_attrs: dict, node_storage_path: str) -> ElectronMetadata:
    task_group_id = node_attrs["task_group_id"]
    name = node_attrs["name"]
    executor = node_attrs["metadata"]["executor"]
    executor_data = node_attrs["metadata"]["executor_data"]

    # Optional
    status = node_attrs.get("status", RESULT_STATUS.NEW_OBJECT)

    start_time = node_attrs.get("start_time", None)
    if start_time:
        start_time = start_time.isoformat()

    end_time = node_attrs.get("end_time", None)
    if end_time:
        end_time = end_time.isoformat()

    return ElectronMetadata(
        task_group_id=task_group_id,
        name=name,
        executor=executor,
        executor_data=executor_data,
        status=str(status),
        start_time=start_time,
        end_time=end_time,
    )


def _deserialize_node_metadata(meta: ElectronMetadata) -> dict:
    return {
        "task_group_id": meta.task_group_id,
        "name": meta.name,
        "status": Status(meta.status),
        "start_time": meta.start_time,
        "end_time": meta.end_time,
        "sub_dispatch_id": meta.sub_dispatch_id,
        "metadata": {
            "executor": meta.executor,
            "executor_data": meta.executor_data,
        },
    }


def _serialize_node_assets(node_attrs: dict, node_storage_path: str) -> ElectronAssets:
    function = node_attrs["function"]
    function_asset = save_asset(
        function,
        ASSET_TYPES["function"],
        node_storage_path,
        ASSET_FILENAME_MAP["function"],
    )

    function_string = node_attrs.get("function_string", "")
    function_string_asset = save_asset(
        function_string,
        ASSET_TYPES["function_string"],
        node_storage_path,
        ASSET_FILENAME_MAP["function_string"],
    )

    node_value = node_attrs.get("value", None)
    value_asset = save_asset(
        node_value,
        ASSET_TYPES["value"],
        node_storage_path,
        ASSET_FILENAME_MAP["value"],
    )

    node_output = node_attrs.get("output", None)
    output_asset = save_asset(
        node_output,
        ASSET_TYPES["output"],
        node_storage_path,
        ASSET_FILENAME_MAP["output"],
    )

    node_error = node_attrs.get("error", "")
    error_asset = save_asset(
        node_error,
        ASSET_TYPES["error"],
        node_storage_path,
        ASSET_FILENAME_MAP["error"],
    )

    node_stdout = node_attrs.get("stdout", "")
    stdout_asset = save_asset(
        node_stdout,
        ASSET_TYPES["stdout"],
        node_storage_path,
        ASSET_FILENAME_MAP["stdout"],
    )
    node_stderr = node_attrs.get("stderr", "")
    stderr_asset = save_asset(
        node_error,
        ASSET_TYPES["stderr"],
        node_storage_path,
        ASSET_FILENAME_MAP["stderr"],
    )

    deps = node_attrs["metadata"]["deps"]
    deps_asset = save_asset(
        deps, ASSET_TYPES["deps"], node_storage_path, ASSET_FILENAME_MAP["deps"]
    )

    call_before = node_attrs["metadata"]["call_before"]
    call_before_asset = save_asset(
        call_before,
        ASSET_TYPES["call_before"],
        node_storage_path,
        ASSET_FILENAME_MAP["call_before"],
    )

    call_after = node_attrs["metadata"]["call_after"]
    call_after_asset = save_asset(
        call_after,
        ASSET_TYPES["call_after"],
        node_storage_path,
        ASSET_FILENAME_MAP["call_after"],
    )

    return ElectronAssets(
        function=function_asset,
        function_string=function_string_asset,
        value=value_asset,
        output=output_asset,
        deps=deps_asset,
        call_before=call_before_asset,
        call_after=call_after_asset,
        stdout=stdout_asset,
        stderr=stderr_asset,
        error=error_asset,
    )


def _deserialize_node_assets(ea: ElectronAssets) -> dict:
    function = load_asset(ea.function, AssetType.TRANSPORTABLE)
    function_string = load_asset(ea.function_string, AssetType.TEXT)
    value = load_asset(ea.value, AssetType.TRANSPORTABLE)
    output = load_asset(ea.output, AssetType.TRANSPORTABLE)
    deps = load_asset(ea.deps, AssetType.OBJECT)
    call_before = load_asset(ea.call_before, AssetType.OBJECT)
    call_after = load_asset(ea.call_after, AssetType.OBJECT)

    stdout = load_asset(ea.stdout, AssetType.TEXT)
    stderr = load_asset(ea.stderr, AssetType.TEXT)
    error = load_asset(ea.error, AssetType.TEXT)

    return {
        "function": function,
        "function_string": function_string,
        "value": value,
        "output": output,
        "stdout": stdout,
        "stderr": stderr,
        "error": error,
        "metadata": {
            "deps": deps,
            "call_before": call_before,
            "call_after": call_after,
        },
    }


def serialize_node(node_id: int, node_attrs: dict, node_storage_path) -> ElectronSchema:
    meta = _serialize_node_metadata(node_attrs, node_storage_path)
    assets = _serialize_node_assets(node_attrs, node_storage_path)
    return ElectronSchema(id=node_id, metadata=meta, assets=assets)


def deserialize_node(e: ElectronSchema, metadata_only: bool = False) -> dict:
    node_attrs = _deserialize_node_metadata(e.metadata)
    node_assets = _deserialize_node_assets(e.assets)

    asset_metadata = node_assets.pop("metadata")

    # merge "metadata" attrs
    node_attrs.update(node_assets)
    node_attrs["metadata"].update(asset_metadata)

    return {"id": e.id, "attrs": node_attrs}



# Copyright 2023 Agnostiq Inc.


from contextlib import contextmanager
from typing import Any, Dict, List, Optional

import pennylane as qml
from pydantic import BaseModel  # pylint: disable=no-name-in-module

from ..core.device import QEDevice
from ..core.future_result import QNodeFutureResult
from ..executors.base import CloudQExecutor


class QNodeSpecs(BaseModel):
    """
    A container for the specifications of a QNode.
    """
    gate_sizes: Dict[str, int]
    gate_types: Dict[str, int]
    num_operations: int
    num_observables: int
    num_diagonalizing_gates: int
    num_used_wires: int
    depth: int
    num_trainable_params: int = None
    num_device_wires: int
    device_name: str
    diff_method: Optional[str]
    expansion_strategy: str
    gradient_options: Dict[str, int]
    interface: Optional[str]
    gradient_fn: Optional[str]
    num_gradient_executions: Any = 0
    num_parameter_shift_executions: int = None


class QElectronInfo(BaseModel):
    """
    A container for related settings used by the wrapping QElectron.
    """
    name: str
    description: str = None
    qnode_device_shots: Optional[int]  # optional default for execution devices
    num_device_wires: int  # this can not be reliably inferred from tapes alone
    pennylane_active_return: bool  # client-side status of `pennylane.active_return()`


_GRADIENT_ACCESS_MAXES = {
    "parameter-shift": 2,
}


class QNodeQE(qml.QNode):
    """
    A sub-type of Pennylane's QNode that integrates Covalent QElectrons.

    Attributes:
        original_qnode: The original QNode that was wrapped.
        device: The `QEDevice` instance used for circuit execution instead of the
            original QNode's device.
    """

    def __init__(
        self,
        qnode: qml.QNode,
        executors: List[CloudQExecutor],
        qelectron_info: QElectronInfo,
    ):
        """
        Initialize a `QNodeQE` instance.

        Args:
            qnode: The Pennylane QNode replaced by this object.
            executors: A list of executors to use for circuit execution.
            qelectron_info: Settings related to the original QNode.
        """

        self.original_qnode = qnode

        # Create a new device for every QNodeQE instance
        qe_device = QEDevice(
            wires=qnode.device.num_wires,
            shots=qnode.device.shots,
            executors=executors,
            qelectron_info=qelectron_info,
        )

        super().__init__(
            func=qnode.func,
            device=qe_device,
            interface=qnode.interface,
            diff_method=qnode.diff_method,
            expansion_strategy=qnode.expansion_strategy,
            max_expansion=qnode.max_expansion,
        )

        self._gradient_fn = None
        self._override_gradient_fn = None
        self._gradient_access_counter = 0

        # Update `execute_kwargs` such that `qe_device.batch_execute` will be called
        # to obtain the circuits result, and `qe_device.gradients` will be called
        # to obtain the gradients.
        if self.device.qelectron_info.pennylane_active_return:
            self.execute_kwargs.update(grad_on_execution=False)
        else:
            self.execute_kwargs.update(mode="backward")

    @contextmanager
    def mark_call_async(self):
        """
        Activates async execution mode for this instance.
        """

        self.device._async_run = True
        try:
            yield
        finally:
            self.device._async_run = False
            self.device._batch_id = None

    def run_later(self, *args, **kwargs) -> QNodeFutureResult:
        """
        Run the QNode asynchronously.

        Args:
            *args: Positional arguments to pass to the QNode.
            **kwargs: Keyword arguments to pass to the QNode.

        Returns:
            FutureResult: A wrapper object for the async result of running the QNode.
        """

        with self.mark_call_async():
            self(*args, **kwargs)
            batch_id = self.device._batch_id

        return QNodeFutureResult(batch_id, self.original_qnode, self.tape)

    def __call__(self, *args, **kwargs):

        self.device.qnode_specs = QNodeSpecs(**qml.specs(self)(*args, **kwargs))
        with self.override_gradient_fn("device"):
            return super().__call__(*args, **kwargs)

    @contextmanager
    def override_gradient_fn(self, gradient_fn):
        """
        Set the `_override_gradient_fn` attribute to enable custom `gradient_fn`
        property behavior.
        """
        self._override_gradient_fn = gradient_fn
        try:
            yield
        finally:
            self._override_gradient_fn = None

    @property
    def gradient_access_max(self):
        """
        Return the maximum number of times the `gradient_fn` property can be
        accessed before the overridden value is returned and the counter is reset.
        """
        return _GRADIENT_ACCESS_MAXES.get(self.diff_method, -1)

    @property
    def gradient_fn(self):
        """
        Override the `gradient_fn` attribute to return custom value (as set by
        `override_gradient_fn`) every second time the property is accessed.
        """
        if (
            self._override_gradient_fn and self._gradient_access_counter >= self.gradient_access_max
        ):
            self.reset_gradient_counter()
            return self._override_gradient_fn

        # Increment access counter.
        self._gradient_access_counter += 1
        return self._gradient_fn

    @gradient_fn.setter
    def gradient_fn(self, fn):
        # Set attribute and reset access counter.
        self._gradient_fn = fn
        self.reset_gradient_counter()

    def reset_gradient_counter(self):
        """
        Reset the gradient access counter to 0.
        """
        self._gradient_access_counter = 0



# Copyright 2023 Agnostiq Inc.

"""Functions to convert lattice -> LatticeSchema"""


from covalent._results_manager.result import Result
from covalent._shared_files.util_classes import Status

from ..shared.schemas.result import ASSET_FILENAME_MAP, ResultAssets, ResultMetadata, ResultSchema
from .common import AssetType, load_asset, save_asset
from .lattice import deserialize_lattice, serialize_lattice

ASSET_TYPES = {
    "inputs": AssetType.OBJECT,
    "error": AssetType.TEXT,
    "result": AssetType.TRANSPORTABLE,
}


def _serialize_result_metadata(res: Result) -> ResultMetadata:
    return ResultMetadata(
        dispatch_id=res._dispatch_id,
        root_dispatch_id=res._root_dispatch_id,
        status=str(res._status),
        start_time=res._start_time,
        end_time=res._end_time,
    )


def _deserialize_result_metadata(meta: ResultMetadata) -> dict:
    return {
        "_dispatch_id": meta.dispatch_id,
        "_root_dispatch_id": meta.root_dispatch_id,
        "_status": Status(meta.status),
        "_start_time": meta.start_time,
        "_end_time": meta.end_time,
    }


def _serialize_result_assets(res: Result, storage_path: str) -> ResultAssets:
    # NOTE: We can avoid pickling here since the UI actually consumes only the string representation
    inputs_asset = save_asset(
        res._inputs, ASSET_TYPES["inputs"], storage_path, ASSET_FILENAME_MAP["inputs"]
    )

    error_asset = save_asset(
        res._error, ASSET_TYPES["error"], storage_path, ASSET_FILENAME_MAP["error"]
    )
    result_asset = save_asset(
        res._result,
        ASSET_TYPES["result"],
        storage_path,
        ASSET_FILENAME_MAP["result"],
    )
    return ResultAssets(inputs=inputs_asset, result=result_asset, error=error_asset)


def _deserialize_result_assets(assets: ResultAssets) -> dict:
    error = load_asset(assets.error, AssetType.TEXT)
    result = load_asset(assets.result, AssetType.TRANSPORTABLE)
    inputs = load_asset(assets.inputs, AssetType.OBJECT)
    return {"_result": result, "_error": error, "_inputs": inputs}


def serialize_result(res: Result, storage_path: str) -> ResultSchema:
    meta = _serialize_result_metadata(res)
    assets = _serialize_result_assets(res, storage_path)
    lat = serialize_lattice(res.lattice, storage_path)
    return ResultSchema(metadata=meta, assets=assets, lattice=lat)


def deserialize_result(res: ResultSchema) -> Result:
    dispatch_id = res.metadata.dispatch_id
    lat = deserialize_lattice(res.lattice)
    result_object = Result(lat, dispatch_id)
    attrs = _deserialize_result_metadata(res.metadata)
    assets = _deserialize_result_assets(res.assets)

    attrs.update(assets)
    result_object.__dict__.update(attrs)
    return result_object


# Functions to preprocess manifest for submission


def strip_local_uris(res: ResultSchema) -> ResultSchema:
    # Create a copy with the local uris removed for submission
    manifest = res.copy(deep=True).dict()

    # Strip workflow asset uris:
    dispatch_assets = manifest["assets"]
    for _, asset in dispatch_assets.items():
        asset["uri"] = ""

    lattice = manifest["lattice"]
    lattice_assets = lattice["assets"]
    for _, asset in lattice_assets.items():
        asset["uri"] = ""

    # Node assets
    tg = lattice["transport_graph"]

    nodes = tg["nodes"]
    for node in nodes:
        node_assets = node["assets"]
        for _, asset in node_assets.items():
            asset["uri"] = ""

    return ResultSchema.parse_obj(manifest)


# Functions to postprocess response


def merge_response_manifest(res: ResultSchema, response: ResultSchema) -> ResultSchema:
    res.metadata.dispatch_id = response.metadata.dispatch_id
    res.metadata.root_dispatch_id = response.metadata.root_dispatch_id

    # Strip workflow asset uris:
    dispatch_assets = response.assets
    for key, asset in res.assets:
        remote_asset = getattr(dispatch_assets, key)
        asset.remote_uri = remote_asset.remote_uri

    lattice = response.lattice
    lattice_assets = lattice.assets
    for key, asset in res.lattice.assets:
        remote_asset = getattr(lattice_assets, key)
        asset.remote_uri = remote_asset.remote_uri

    # Node assets
    tg = lattice.transport_graph

    # Sort returned nodes b/c task packing may reorder nodes
    tg.nodes.sort(key=lambda x: x.id)
    nodes = res.lattice.transport_graph.nodes

    for i, node in enumerate(nodes):
        returned_node = tg.nodes[i]
        returned_node_assets = returned_node.assets
        for key, asset in node.assets:
            remote_asset = getattr(returned_node_assets, key)
            asset.remote_uri = remote_asset.remote_uri
    return res


def extract_assets(manifest: dict) -> list:
    assets = []

    # workflow-level assets
    dispatch_assets = manifest["assets"]
    for key, asset in dispatch_assets.items():
        assets.append(asset)

    lattice = manifest["lattice"]
    lattice_assets = lattice["assets"]
    for key, asset in lattice_assets.items():
        assets.append(asset)

    # Node assets
    tg = lattice["transport_graph"]

    nodes = tg["nodes"]
    for node in nodes:
        node_assets = node["assets"]
        for key, asset in node_assets.items():
            assets.append(asset)
    return assets



# Copyright 2023 Agnostiq Inc.


from abc import ABC, abstractmethod, abstractproperty

class BaseQClient(ABC):
    @abstractmethod
    def submit(self, qscripts, executors, qelectron_info, qnode_specs):
        raise NotImplementedError

    @abstractmethod
    def get_results(self, batch_id):
        raise NotImplementedError

    @abstractproperty
    def selector(self):
        raise NotImplementedError

    @abstractproperty
    def database(self):
        raise NotImplementedError

    # The following methods are abstract because the qserver
    # is expecting serialized inputs and will be sending
    # back serialized outputs, thus even if these methods
    # essentially just pass through, they are still to be 
    # implemented by the child class and should use the same 
    # seriliazing/deserializing methods as are being used by the equivalent qserver.

    @abstractmethod
    def serialize(self, obj):
        raise NotImplementedError

    @abstractmethod
    def deserialize(self, ser_obj):
        raise NotImplementedError



# Copyright 2023 Agnostiq Inc.

"""Unit tests for dispatch helpers"""


import covalent as ct

from covalent_cloud.dispatch_management.helpers import inject_parameter_outputs


def test_parameter_outputs():
    """Test inject_parameter_outputs"""

    @ct.electron
    def task(x):
        return x**2

    @ct.lattice
    def workflow(x):
        return task(x)

    workflow.build_graph(3)

    inject_parameter_outputs(workflow)

    tg = workflow.transport_graph
    param_output = tg.get_node_value(1, "output")
    param_value = tg.get_node_value(1, "output")

    assert param_output.get_deserialized() == param_value.get_deserialized()



# Copyright 2023 Agnostiq Inc.


"""Models for /api/v1/resultv2 endpoints"""

from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from .asset import AssetSchema
from .lattice import (
    LATTICE_ERROR_FILENAME,
    LATTICE_INPUTS_FILENAME,
    LATTICE_RESULTS_FILENAME,
    LatticeSchema,
)

METADATA_KEYS = {
    "start_time",
    "end_time",
    "dispatch_id",
    "root_dispatch_id",
    "status",
    "num_nodes",
}


ASSET_KEYS = {
    "inputs",
    "result",
    "error",
}


ASSET_FILENAME_MAP = {
    "inputs": LATTICE_INPUTS_FILENAME,
    "result": LATTICE_RESULTS_FILENAME,
    "error": LATTICE_ERROR_FILENAME,
}


class ResultMetadata(BaseModel):
    dispatch_id: str
    root_dispatch_id: str
    status: str
    start_time: Optional[datetime]
    end_time: Optional[datetime]


class ResultAssets(BaseModel):
    inputs: AssetSchema
    result: AssetSchema
    error: AssetSchema


class ResultSchema(BaseModel):
    metadata: ResultMetadata
    assets: ResultAssets
    lattice: LatticeSchema



# Copyright 2023 Agnostiq Inc.

import os
from pathlib import Path
from typing import Union

from pydantic import BaseModel, BaseSettings


class AuthSettings(BaseModel):
    """
    Authentication settings.

    token: Authentication token.
    api_key: API key.
    config_file: Path to the config file. Defaults to "~/.config/covalent_cloud/credentials.toml".
    config_file_section: Section in the config file. Defaults to "auth".
    cofig_file_token_keyname: Keyname for the token in the config file. Defaults to "token".
    config_file_api_key_keyname: Keyname for the API key in the config file. Defaults to "api_key".

    """

    token: str = ""  # AUTH__TOKEN env
    api_key: str = ""  # api_key
    config_file: str = str(
        Path(os.environ.get("HOME")) / Path(".config/covalent_cloud") / Path("credentials.toml")
    )  # AUTH__CONFIG_FILE env
    config_file_section: str = "auth"
    cofig_file_token_keyname: str = "token"
    config_file_api_key_keyname: str = "api_key"


class Settings(BaseSettings):
    """
    Settings for the Covalent Cloud.

    auth: Authentication settings.
    dispatcher_uri: URI for the dispatcher. Defaults to "https://api.covalent.xyz".
    dispatcher_port: Port for the dispatcher. Defaults to None.
    dispatch_cache_dir: Directory for the dispatch cache. Defaults to "~/.cache/covalent/dispatches".
    results_dir: Directory for the results. Defaults to "~/.cache/covalent/results".
    validate_executors: Whether to validate executors. Defaults to True.

    """

    auth: AuthSettings = AuthSettings()

    dispatcher_uri: str = "https://api.covalent.xyz"
    dispatcher_port: Union[int, None] = None

    dispatch_cache_dir: str = os.environ["HOME"] + "/.cache" + "/covalent/dispatches"
    results_dir: str = os.environ["HOME"] + "/.cache" + "/covalent/results"

    validate_executors: bool = True

    class Config:
        env_prefix = "COVALENT_CLOUD_"
        env_nested_delimiter = "__"


settings = Settings()



# Copyright 2023 Agnostiq Inc.

"""Software environment management module."""

import json
import tempfile
from typing import Dict, List, Optional, Tuple, Union

import requests
import yaml

from covalent_cloud import get_client

from ..shared.classes.exceptions import CovalentAPIKeyError, CovalentGenericAPIError

COVALENT_CLOUD_URL = "fake.fake"
API_KEY = "fake"  # pragma: allowlist secret


def get_pip_pkgs(pip: Union[str, List[str]]) -> List[str]:
    """
    Unpacks the pip packages in the requirements.txt file and combines it into a list of required pip packages.

    """

    if not isinstance(pip, list):
        pip = [pip]

    pip_pkgs = []
    for pkg in pip:
        if pkg.endswith(".txt"):
            with open(pkg, "r") as f:
                pip_pkgs += f.read().splitlines()
        else:
            pip_pkgs.append(pkg)
    return pip_pkgs


def unpack_conda_pkgs(
    conda: Union[str, List[str], Dict[str, List[str]]]
) -> Tuple[List[str], List[str]]:
    """
    Unpacks the conda packages in the environment.yml file and combines it into a dictionary of required conda packages.

    Returns:
        channels, dependencies: channels and dependencies according to the conda environment.yml file. Note that these terms are chosen according to the conda nomenclature.

    """

    channels, dependencies = [], []

    if isinstance(conda, dict):
        channels = conda.get("channels", [])
        dependencies = conda.get("dependencies", [])

    elif isinstance(conda, str):
        if conda.endswith(".yml"):
            with open(conda, "r") as f:
                parsed_conda_env_yaml = yaml.safe_load(f)

            channels = parsed_conda_env_yaml.get("channels", [])
            dependencies = parsed_conda_env_yaml.get("dependencies", [])

    elif isinstance(conda, list):
        dependencies = conda

    return channels, dependencies


def create_env(
    name: str,
    pip: Union[str, List[str]],
    conda: Union[str, List[str], Dict[str, List[str]]],
    variables: Optional[List] = None,
) -> None:
    """
    Sends the create request to the Covalent Cloud server with the environment dependency list.

    Args:
        name: Identifier/name for the software environment.

        pip: Python packages to be installed in the environment using pip. This value can be a string `requirements.txt` and/or a list of packages. Note, that if it's a list, it's possible that one of the values is the string `requirements.txt`. In case a `requirements.txt` is passed, it will be parsed into a list of packages and combined with the list of packages passed.`

        conda: List of packages to be installed in the environment using conda. This value can either be a list of packages, a filepath to `environment.yml`. It could also be a dictionary with channels, dependencies, and (optionally) variables as keys, and a list of strings as their values. For example:

            conda={
                        "channels": ["conda-forge", "defaults"],
                        "dependencies": ["numpy=1.21.*", "xarray=0.15.1"],
                        "variables": [{'name': 'mock-variable-1', 'value': '1', 'sensitive': True}]
            }

        Whatever is passed, it will be parsed into a dictionary as shown above and sent as JSON to the Covalent Cloud server. If a list of packages is provided, they will be installed using the default conda channel.

    Returns:
        None

    Examples:
        Create an environment with a list of packages:
            >>> create_env("test-env", ["typing"], ["numpy=1.21.*", "xarray=0.15.1"])

        Create an environment with a filepath to `environment.yml`:
            >>> create_env("test-env", "requirements.txt", "environment.yml")

        Create an environment with a dictionary of channels, dependencies, and variables:
            >>> create_env("test-env", "requirements.txt", {"channels": ["conda-forge", "defaults"], "dependencies": ["numpy=1.21.*", "xarray=0.15.1"], "variables": [{'name': 'mock-variable-1', 'value': '1', 'sensitive': True}]})

    Note:
        In case of a conflict of package between pip and conda, pip will take precedence and the conda one will be ignored.

    """

    if variables is None:
        variables = []

    pip_pkgs = get_pip_pkgs(pip)
    channels, dependencies = unpack_conda_pkgs(conda)

    dependencies.append({"pip": pip_pkgs})
    yaml_template = {
        "name": name,
        "channels": channels,
        "dependencies": dependencies,
    }

    response_body = None

    with tempfile.NamedTemporaryFile(mode="w", suffix=".yml", delete=False) as definition_file:

        yaml.dump(yaml_template, definition_file, default_flow_style=False)

        client = get_client()

        # Open a separate reader in binary mode per Requests doc
        try:
            with open(definition_file.name, "rb") as def_file_reader:
                response = client.post(
                    "/api/v2/envs",
                    {
                        "files": {"definition": def_file_reader},
                        "data": {"name": name, "variables": json.dumps(variables)},
                    },
                )
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 400:
                print("Environment Already Exists.")
                return
            elif (
                e.response.status_code == 401 and e.response.json()["code"] == "auth/unauthorized"
            ):
                CovalentAPIKeyError(
                    message="A valid API key is required to create an environment.",
                    code=e.response.json()["code"],
                ).rich_print(level="error")
                return
            else:
                raise CovalentGenericAPIError(error=e)

        response_body = response.json()

    print(f"Name: {response_body['name']}")
    print(f"Status: {response_body['status']}")
    print(f"Estimated Time: {response_body['estimated_time']} seconds")

    print("Environment file contains:")
    print("==========================")
    print(yaml.dump(yaml_template, default_flow_style=False))


def delete_env(env_name: str) -> None:
    """
    Sends the delete request to the Covalent Cloud server with the environment name to be deleted.

    Args:
        env_name: Identifier/name for the software environment.

    Returns:
        None

    """

    params = {"env_name": env_name, "api_key": API_KEY}

    response = requests.delete(f"{COVALENT_CLOUD_URL}/delete_env", params=params)
    response.raise_for_status()

    response_body = {
        "name": "Deleting",
        "description": "Environment's deletion is in progress",
    }

    print(f"Status: {response_body['name']}")
    print(f"Description: {response_body['description']}")



# Copyright 2023 Agnostiq Inc.

"""Module for Covalent Cloud dispatching and related functionalities."""


import lzma
from copy import deepcopy
from functools import wraps
from typing import Any, Callable

from covalent._results_manager.result import Result
from covalent._workflow.lattice import Lattice
from covalent._workflow.transport import TransportableObject

from covalent_cloud.shared.classes.exceptions import CovalentAPIKeyError

from ..shared.classes.settings import Settings, settings
from . import results_manager as rm
from .helpers import register, start, validate_executors

API_KEY = "fake"  # pragma: allowlist secret


def _compressed_transportable(data: Any) -> bytes:
    """Applies LZMA compression to a transportable object serialized into bytes."""
    return lzma.compress(TransportableObject(data).serialize())


def dispatch(
    orig_lattice: Lattice,
    settings: Settings = settings,
) -> Callable:
    """
    Dispatches a Covalent workflow to the Covalent Cloud and returns the assigned dispatch ID.
    The dispatch function takes a Covalent workflow, also called a lattice, and sends it to the Covalent Cloud Server for execution. Once dispatched, the workflow runs on a the cloud and can be monitored using the dispatch ID in the application. The dispatch function returns a wrapper function that takes the inputs of the workflow as arguments. This wrapper function can be called to execute the workflow.

    Args:
        orig_lattice: The Covalent workflow to send to the cloud.
        settings: The settings object to use. If None, the default settings will be used.

    Returns:
        A wrapper function which takes the inputs of the workflow as arguments.

    Examples:

        # define a simple lattice
        import covalent_cloud as ctc
        import covalent as ct
        
        @ct.lattice
        def my_lattice(a: int, b: int) -> int:
            return a + b

        # dispatch the lattice and get the assigned dispatch ID
        dispatch_id = ctc.dispatch(my_lattice)(2, 3)
    """

    dispatcher_addr = settings.dispatcher_uri

    @wraps(orig_lattice)
    def wrapper(*args, **kwargs) -> str:
        """
        Send the lattice to the dispatcher server and return
        the assigned dispatch id.

        Args:
            *args: The inputs of the workflow.
            **kwargs: The keyword arguments of the workflow.

        Returns:
            The dispatch id of the workflow.

        """

        try:
            lattice = deepcopy(orig_lattice)
            lattice.build_graph(*args, **kwargs)

            # Check that only CloudExecutors are specified.
            if settings.validate_executors and not validate_executors(lattice):
                raise ValueError("One or more electrons have invalid executors.")

            dispatch_id = register(lattice, settings)(*args, **kwargs)
            return start(dispatch_id)

        except CovalentAPIKeyError as e:
            e.rich_print(level="error")

    return wrapper


def get_result(
    dispatch_id: str,
    wait: bool = False,
    settings: Settings = settings,
    *,
    status_only: bool = False,
) -> Result:
    """
    Gets the result of a Covalent workflow that has been dispatched to the cloud.

    This function retrieves the result of a dispatched Covalent workflow that has been executed on the cloud. The function takes the dispatch ID of the workflow and retrieves the results from the cloud. The result is returned as a `Result` object that contains the status of the workflow, the final result of the lattice, and any error messages that may have occurred during execution.

    Args:
        dispatch_id: The dispatch ID assigned to the workflow.
        wait: Controls how long the function waits for the server to return a result. If False, the function will not wait and will return the current status of the workflow. If True, the function will wait for the result to finish and keep retrying until the result is available.
        status_only: If True, only retrieves the status of the workflow and not the full result.
        settings: The Covalent settings to use for the request.

    Returns:
        A `Result` object that contains the status of the workflow, the final result of the lattice, and any error messages that may have occurred during execution.

    Examples:

    # define a simple lattice
    import covalent_cloud as ctc
    import covalent as ct
    
    @ct.lattice
    def my_lattice(a: int, b: int) -> int:
        return a + b

    # dispatch the lattice and get the assigned dispatch ID
    dispatch_id = ctc.dispatch(my_lattice)(2, 3)

    # get the result of the dispatched lattice
    result = ctc.get_result(dispatch_id, wait=True)

    # print the final result of the lattice
    print(result.result)
    """

    return rm.get_result(
        dispatch_id=dispatch_id,
        wait=wait,
        settings=settings,
        status_only=status_only,
    )



# Copyright 2023 Agnostiq Inc.


from pennylane import QubitDevice
from pennylane import numpy as np
from pennylane.devices.default_qubit import DefaultQubit
from pennylane.gradients import param_shift

from ..middleware.core import middleware


class QEDevice(QubitDevice):

    """
    Custom Pennylane device that batch executes circuits by submitting them to the middleware
    """

    name = "QEDevice"
    short_name = "qe_device"
    pennylane_requires = ">=0.29.1"
    version = "0.0.1"
    author = "aq"

    operations = DefaultQubit.operations
    observables = DefaultQubit.observables

    def __init__(
        self,
        wires=1,
        shots=None,
        *,
        r_dtype=np.float64,
        c_dtype=np.complex128,
        analytic=None,
        executors=None,
        qelectron_info=None,
    ):
        super().__init__(wires, shots, r_dtype=r_dtype, c_dtype=c_dtype, analytic=analytic)

        self._async_run = False
        self._batch_id = None
        self.executors = executors
        self.qelectron_info = qelectron_info

        # This will be set when the QNodeQE
        # is called with args and kwargs
        self.qnode_specs = None

    def batch_execute(self, circuits):

        # Async submit all circuits to middleware.
        batch_id = middleware.run_circuits_async(
            circuits, self.executors, self.qelectron_info, self.qnode_specs
        )

        # Relevant when `run_later` is used to run the circuit.
        # We will retrieve the result later using the future object and the batch_id.
        if self._async_run:
            self._batch_id = batch_id
            # Return a dummy result
            return [self._asarray([1] * c.output_dim) for c in circuits]

        # Otherwise, get the results from the middleware
        results = middleware.get_results(batch_id)

        return results

    def gradients(self, circuits, method="jacobian", **kwargs):

        # Obtain gradient tapes and processing functions.
        grad_tapes_tuple, processing_fns = zip(
            *(param_shift(circuit, **kwargs) for circuit in circuits)
        )

        # Flatten the list of gradient tapes.
        nonempty_grad_tapes_list = []
        for grad_tapes in grad_tapes_tuple:
            nonempty_grad_tapes_list.extend(grad_tapes)

        # Skip gradient computation for vectorized calls.
        # Also require nonempty list of gradient tapes.
        scalar_input = all(not circuit.batch_size for circuit in circuits)
        if not (scalar_input and nonempty_grad_tapes_list):
            return []

        # Submit gradient circuits to middleware.
        nonempty_grad_res = self.batch_execute(nonempty_grad_tapes_list)

        # Unpack the nonempty gradient result into the expected shape.
        grad_res = self._unpack_gradient_result(nonempty_grad_res, grad_tapes_tuple)

        # Compute parameter-shift jacobians.
        jacs = [fn(r) for r, fn in zip(grad_res, processing_fns)]

        return jacs

    def _unpack_gradient_result(self, nonzero_grads, grad_tapes_tuple):
        """
        Unpack the gradient tapes' execution results into a list containing arrays
        and lists of arrays. Insert zero-arrays for empty gradient tapes.
        """

        grad_res = []
        zero_arr = self._asarray(0.)
        nonempty_idx = 0
        for grad_tapes in grad_tapes_tuple:
            if not grad_tapes:
                grad_res.append(zero_arr)
            else:
                sub_res = []
                for _ in grad_tapes:
                    sub_res.append(nonzero_grads[nonempty_idx])
                    nonempty_idx += 1

                sub_res = sub_res[0] if len(sub_res) == 1 else sub_res
                grad_res.append(sub_res)

        return grad_res

    def apply(self, *args, **kwargs):
        # Dummy implementation of abstractmethod on `QubitDevice`
        pass

    @classmethod
    def capabilities(cls):
        capabilities = super().capabilities().copy()
        capabilities.update(supports_broadcasting=True)
        return capabilities



# Copyright 2023 Agnostiq Inc.

import hashlib
import json
from enum import Enum
from pathlib import Path
from typing import Any

import cloudpickle

from ..shared.schemas.asset import AssetSchema

CHECKSUM_ALGORITHM = "sha"


class AssetType(Enum):
    OBJECT = 0
    TRANSPORTABLE = 1
    JSONABLE = 2
    TEXT = 3


def _serialize_asset(data: Any, data_type: AssetType) -> bytes:
    if data_type == AssetType.OBJECT:
        return cloudpickle.dumps(data)
    elif data_type == AssetType.TRANSPORTABLE:
        return cloudpickle.dumps(data)
    elif data_type == AssetType.JSONABLE:
        return json.dumps(data)
    elif data_type == AssetType.TEXT:
        if data is None:
            data = ""
        return data.encode("utf-8")
    else:
        raise TypeError(f"Unsupported data type {type(data)}")


def _deserialize_asset(data: bytes, data_type: AssetType) -> Any:
    if data_type == AssetType.OBJECT:
        return cloudpickle.loads(data)
    elif data_type == AssetType.TRANSPORTABLE:
        return cloudpickle.loads(data)
    elif data_type == AssetType.JSONABLE:
        return json.loads(data)
    elif data_type == AssetType.TEXT:
        return data.decode("utf-8")
    else:
        raise TypeError("Unsupported data type")


def _sha1_asset(data: bytes) -> str:
    return hashlib.sha1(data).hexdigest()


def save_asset(data: Any, data_type: AssetType, storage_path: str, filename: str) -> AssetSchema:
    scheme = "file"

    serialized = _serialize_asset(data, data_type)
    digest = _sha1_asset(serialized)
    path = Path(storage_path) / filename
    path = path.resolve()
    with open(path, "wb") as f:
        f.write(serialized)
    uri = f"{scheme}://{path}"
    return AssetSchema(digest_alg=CHECKSUM_ALGORITHM, digest=digest, size=len(serialized), uri=uri)


def load_asset(asset_meta: AssetSchema, data_type: AssetType) -> Any:
    scheme_prefix = "file://"
    uri = asset_meta.uri

    if not uri:
        return None

    if uri.startswith(scheme_prefix):
        path = uri[len(scheme_prefix) :]
    else:
        path = uri
    with open(path, "rb") as f:
        data = f.read()
    return _deserialize_asset(data, data_type)



# Copyright 2023 Agnostiq Inc.

"""Unit tests for dispatching and related interface functions."""

from unittest.mock import MagicMock

import covalent as ct
import pytest

from covalent_cloud.dispatch_management.interface_functions import dispatch, get_result
from covalent_cloud.dispatch_management.results_manager import Result

COVALENT_CLOUD_URL = "http://localhost:48008"
MOCK_DISPATCH_ID = "mock-dispatch-id"
MOCK_STATUS = "mock-status"
MOCK_ARGS = [1, 2]
MOCK_KWARGS = {"a": 1, "b": 2}
MOCK_TASKS_LIST = ["task1", "task2"]
MOCK_COMPRESSED_TRANSPORTABLE_DATA = "mock-compressed-transportable-data"


def test_dispatch(mocker):
    """Test the workflow dispatching interface function."""

    @ct.lattice(executor="cloud", workflow_executor="cloud")
    @ct.electron(executor="cloud")
    def test_workflow(a, b):
        return a + b

    mock_dispatch_id = "mock-dispatch-id"
    mock_wrapper = MagicMock(return_value=mock_dispatch_id)
    mock_register = mocker.patch(
        "covalent_cloud.dispatch_management.interface_functions.register",
        return_value=mock_wrapper,
    )

    mock_start = mocker.patch(
        "covalent_cloud.dispatch_management.interface_functions.start",
        return_value=mock_dispatch_id,
    )

    dispatch_id = dispatch(test_workflow)(2, 3)

    assert dispatch_id == "mock-dispatch-id"
    mock_register.assert_called()
    mock_start.assert_called_with(dispatch_id)


@pytest.mark.parametrize("status_only", [True, False])
def test_get_result(mocker, status_only):
    """Test the workflow result retrieval interface function."""

    @ct.electron
    def task(x):
        return x**2

    @ct.lattice
    def workflow(x):
        return task(x)

    workflow.build_graph(3)
    dispatch_id = f"test_get_result_{status_only}"
    res_obj = Result(workflow, dispatch_id)
    res_obj._initialize_nodes()

    mock_fr = MagicMock()
    mock_rm = MagicMock()
    mock_rm.result_object = res_obj

    mock_get_from_dispatcher = mocker.patch(
        "covalent_cloud.dispatch_management.results_manager._get_result_export_from_dispatcher",
        return_value={"id": dispatch_id, "status": "RUNNING", "result_export": "result"},
    )

    mock_get_rm = mocker.patch(
        "covalent_cloud.dispatch_management.results_manager.get_result_manager",
        return_value=mock_rm,
    )

    mock_from_result_object = mocker.patch(
        "covalent_cloud.dispatch_management.results_manager.FutureResult._from_result_object",
        return_value=mock_fr,
    )

    result = get_result(dispatch_id=dispatch_id, status_only=status_only)
    if status_only:
        mock_get_from_dispatcher.assert_called_once()
        mock_get_rm.assert_not_called()
    else:
        mock_get_rm.assert_called()
        mock_rm._populate_result_object.assert_called()
        assert result == mock_fr
        mock_from_result_object.assert_called_with(res_obj)



# Copyright 2023 Agnostiq Inc.

from typing import Any, Dict, List, Tuple

from pennylane import numpy as np
from pennylane.tape import QuantumScript
from pennylane.transforms import broadcast_expand, map_batch_transform
from pydantic import BaseModel

from ...shared.classes.exceptions import CovalentSDKError
from ..executors import CloudQExecutor
from .qclients import RemoteQClient


class SubmissionRecord(BaseModel):
    """
    A container for useful information about submissions.
    The middleware uses this information to post-process results.
    """

    batch_id: str
    num_qscripts: int
    num_qscripts_expanded: int
    pennylane_active_return: bool

    @property
    def contains_batch(self) -> bool:
        """
        Indicates if the original qscripts were expanded non-trivially.
        """
        return self.num_qscripts != self.num_qscripts_expanded

    @property
    def batch_shape(self) -> Tuple[int, int]:
        """
        Shape of the returned vector.
        """
        return (self.num_qscripts, self.num_qscripts_expanded // self.num_qscripts)


class MiddleWare:
    """
    Middleware object that communicates with the QServer via the Remote QClient.
    """

    def __init__(self) -> None:
        self.qclient = RemoteQClient()

        # Stores records of submission for potential post-processing.
        self._submissions: Dict[str, SubmissionRecord] = {}

    def __new__(cls):
        # Making this a singleton class.
        if not hasattr(cls, 'instance'):
            cls.instance = super(MiddleWare, cls).__new__(cls)
        return cls.instance

    # The following attributes are properties
    # because the qclient might change over time
    # and every time it gets changed, we shouldn't
    # have to set things like:
    # self.database = self.qclient.database
    # Thus, we override the access of these attributes
    # and return/set these dynamically depending upon
    # what the qclient is at that point in time.

    @property
    def selector(self):
        """
        Defers selection to QClient's selector.
        """
        return self.qclient.selector

    @selector.setter
    def selector(self, selector_func):
        self.qclient.selector = selector_func

    @property
    def database(self):
        """
        Points to the QElectron database interface.
        """
        return self.qclient.database

    # Need `QElectronInfo` and `QNodeSpecs` in own module.
    # Can't import these objects here because of circular imports.
    def run_circuits_async(
        self,
        qscripts: List[QuantumScript],
        executors: List[CloudQExecutor],
        qelectron_info,  # NOTE: missing type
        qnode_specs,  # NOTE: missing type
    ) -> str:
        """
        Submit circuits to the QServer via the Remote QClient.

        Args:
            qscripts: List of circuits to be executed.
            executors: :ist of executors to use/
            qelectron_info: QNode device and Global pennylane settings.
            qnode_specs: QNode specifications generated by `qml.specs()`.

        Returns:
            batch_id: Unique ID associated with the submission.
        """
        # Converted batched qscripts into multiple qscripts.
        expanded_qscripts = expand_batch(qscripts)

        # Submit batched qscripts individually.
        batch_id = self.qclient.submit(expanded_qscripts, executors, qelectron_info, qnode_specs)

        # Make record of submission. Post-processing may be required.
        self._record_submission(batch_id, qscripts, expanded_qscripts, qelectron_info)

        return batch_id

    def get_results(self, batch_id: str) -> Any:
        """
        Retrieve results from the QServer via the Remote QClient.

        Args:
            batch_id: Unique ID associated with a submission.

        Returns:
            results: Results of the execution.
        """
        # Get raw results from QServer.
        raw_results = self.qclient.get_results(batch_id)

        # Post-process raw results if needed.
        results = self._post_process(raw_results, batch_id)

        return results

    def _post_process(self, raw_results: List[Any], batch_id: str) -> Any:
        """
        Re-package raw results as necessary. Handle batch submissions.
        """
        # Get submission details.
        submission = self._recall_submission(batch_id)

        if not submission.contains_batch:
            # Post-processing not necessary.
            # Unpack single results. Return multi-return or gradient results as-is.
            return raw_results[0] if len(raw_results) == 1 else raw_results

        # Post-process batched results if needed.
        res = np.asarray(raw_results).squeeze()

        # Adapted from Qiskit device in `devices_base.py` in OS QElectrons.
        shape = submission.batch_shape
        if submission.pennylane_active_return:
            # The case `res.ndim > 1` is for multiple-return circuits.
            return [res] if res.ndim > 1 else list(res.reshape(shape))

        res = res.reshape(-1)
        return [res] if res.ndim > 1 else [res.reshape(shape)]

    def _record_submission(
        self,
        batch_id: str,
        qscripts: List[QuantumScript],
        qscripts_expanded: List[QuantumScript],
        qelectron_info,  # NOTE: missing type
    ) -> None:
        """
        Stores submission details for post-processing.
        """
        if batch_id in self._submissions:
            raise CovalentSDKError(f"Middleware has duplicate batch ID '{batch_id}'")

        self._submissions[batch_id] = SubmissionRecord(
            batch_id=batch_id,
            num_qscripts=len(qscripts),
            num_qscripts_expanded=len(qscripts_expanded),
            pennylane_active_return=qelectron_info.pennylane_active_return,
        )

    def _recall_submission(self, batch_id: str) -> SubmissionRecord:
        """
        Retrieves submission details for post-processing.
        """
        if batch_id not in self._submissions:
            raise CovalentSDKError(f"Middleware missing record for batch ID '{batch_id}'")

        return self._submissions.pop(batch_id)


def expand_batch(circuits: List[QuantumScript]) -> List[QuantumScript]:
    """
    Expand batched circuits into multiple circuits with single parameters.
    """

    if not any(circuit.batch_size for circuit in circuits):
        # No circuit contains a batch. Return the same list of circuits.
        return circuits

    # Expand the circuits.
    expanded_circuits, _ = map_batch_transform(broadcast_expand, circuits)
    return expanded_circuits


middleware = MiddleWare()



# Copyright 2023 Agnostiq Inc.

"""Models for /api/v1/resultv2 endpoints"""

from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from .asset import AssetSchema

ELECTRON_METADATA_KEYS = {
    "task_group_id",
    "name",
    "start_time",
    "end_time",
    "status",
    # electron metadata
    "executor",
    "executor_data",
}

ELECTRON_ASSET_KEYS = {
    "function",
    "function_string",
    "output",
    "value",
    "error",
    "stdout",
    "stderr",
    # electron metadata
    "deps",
    "call_before",
    "call_after",
}

ELECTRON_FUNCTION_FILENAME = "function.pkl"
ELECTRON_FUNCTION_STRING_FILENAME = "function_string.txt"
ELECTRON_VALUE_FILENAME = "value.pkl"
ELECTRON_STDOUT_FILENAME = "stdout.log"
ELECTRON_STDERR_FILENAME = "stderr.log"
ELECTRON_ERROR_FILENAME = "error.log"
ELECTRON_RESULTS_FILENAME = "results.pkl"
ELECTRON_DEPS_FILENAME = "deps.pkl"
ELECTRON_CALL_BEFORE_FILENAME = "call_before.pkl"
ELECTRON_CALL_AFTER_FILENAME = "call_after.pkl"
ELECTRON_STORAGE_TYPE = "file"


ASSET_FILENAME_MAP = {
    "function": ELECTRON_FUNCTION_FILENAME,
    "function_string": ELECTRON_FUNCTION_STRING_FILENAME,
    "value": ELECTRON_VALUE_FILENAME,
    "output": ELECTRON_RESULTS_FILENAME,
    "deps": ELECTRON_DEPS_FILENAME,
    "call_before": ELECTRON_CALL_BEFORE_FILENAME,
    "call_after": ELECTRON_CALL_AFTER_FILENAME,
    "stdout": ELECTRON_STDOUT_FILENAME,
    "stderr": ELECTRON_STDERR_FILENAME,
    "error": ELECTRON_ERROR_FILENAME,
}


class ElectronAssets(BaseModel):
    function: AssetSchema
    function_string: AssetSchema
    value: AssetSchema
    output: AssetSchema
    error: Optional[AssetSchema]
    stdout: Optional[AssetSchema]
    stderr: Optional[AssetSchema]

    # electron_metadata
    deps: AssetSchema
    call_before: AssetSchema
    call_after: AssetSchema


class ElectronMetadata(BaseModel):
    task_group_id: int
    name: str
    executor: str
    executor_data: dict
    sub_dispatch_id: Optional[str]
    status: Optional[str]
    start_time: Optional[datetime]
    end_time: Optional[datetime]


class ElectronSchema(BaseModel):
    id: int
    metadata: ElectronMetadata
    assets: ElectronAssets



# Copyright 2023 Agnostiq Inc.


from rich.console import Console

"""Covalent Cloud SDK Exception module."""


class CovalentSDKError(Exception):
    """Covalent Cloud SDK Base Exception class.

    Attributes:
        message (str): Explanation of the error.
        code (str): String enum representing error analogous to error code.
    """

    def __init__(
        self, message: str = "Generic Error", code: str = "error/generic"
    ) -> None:
        """
        Initializes a new instance of the CovalentSDKError class.

        Args:
            message (str): Explanation of the error.
            code (str): String enum representing error analogous to error code.

        """
        self.message = message
        self.code = code
        super().__init__(f"[{code}] {message}")

    def rich_print(self, level: str = "warning") -> None:
        """Print using the Rich module.

        Args:
            level: The level of the message to print.  Defaults to "warning".

        """
        console = Console()
        if level == "warning":
            console.print(f"[bold yellow1]WARNING: {self.message}[bold yellow1]")
        elif level == "error":
            console.print(f"[bold red1]ERROR: {self.message}[bold red1]")


class CovalentAPIKeyError(CovalentSDKError):
    """Covalent Cloud SDK API Key Error class."""

    def __init__(self, message, code) -> None:
        super().__init__(message, code)


class CovalentGenericAPIError(CovalentSDKError):
    """Covalent Cloud Server Generic API Error class."""

    def __init__(self, error) -> None:
        try:
            error_message = error.response.json()["detail"]
            error_code = error.response.json()["code"]
        except:
            error_message = "Unknown Error"
            error_code = "error/unknown"

        super().__init__(error_message, error_code)
